#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass
from typing import Iterable, Mapping, Annotated

from frozendict import frozendict
from http_tools import ContentType
from init_helpers import Jsonable

from .parameter import ParameterStyle
from .spec_resource import SpecResource
from .spec_ref import SpecRef


@dataclass(frozen=True, slots=True)
class BodyPartDescription(SpecResource):
    content_type: ContentType | tuple[ContentType, ...] | None = None
    headers: Mapping[str, Annotated | type] | None = None
    style: ParameterStyle | None = None
    explode: bool | None = None

    def __init__(
            self,
            content_type: ContentType | Iterable[ContentType] | None = None,
            headers: dict[str, Annotated] | None = None,
            style: ParameterStyle | None = None,
            explode: bool | None = None,
    ) -> None:
        object.__setattr__(
            self, 'content_type',
            content_type if content_type is None or isinstance(content_type, ContentType) else tuple(content_type))
        object.__setattr__(self, 'headers', frozendict(headers) if headers else None)
        object.__setattr__(self, 'style', style)
        object.__setattr__(self, 'explode', explode)

    def get_spec_dependencies(self) -> frozenset['SpecResource']:
        return frozenset(self.headers.values()) if self.headers else frozendict()

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef]) -> frozendict[str, Jsonable]:
        description = {}
        if self.explode is not None:
            description["explode"] = self.explode
        if self.style is not None:
            description["style"] = self.style
        if self.content_type:
            description["content_type"] = ", ".join(self.content_type)
        if self.headers:
            description["headers"] = {k: dependency_to_ref[v] for k, v in self.headers.items()}

        return frozendict(description)



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import collections.abc
import dataclasses
import inspect
import logging
import traceback
import typing
import warnings
from dataclasses import dataclass, field, MISSING
from functools import cached_property
from typing import TypeVar, Any, Callable, Literal, Awaitable, Iterable, Mapping

from async_tools import acall
from dict_caster.extras import is_iterable
from frozendict import frozendict
from http_tools import Answer, HttpStatusCode, ContentType, IncomingRequest, StreamAnswer, JsonableAnswer

from http_tools.answer import BaseAnswer, ExceptionAnswer, ErrorDescription
from init_helpers import raise_if
from init_helpers import NoValue
from more_itertools import first

from openapi_tools.freeze import repr_type
from .endpoint_body import EndpointBody
from .get_response_from_answer import get_response_from_answer
from openapi_tools.spec import SecurityRequirement, CallParameter, SpecParameter, SecurityScheme, RawBodyParameter, \
    Tag, get_securities_arg_names, BodyParameter, RequestBody, MultipartParameter, BodyPartDescription, \
    Unauthorized, Forbidden, ContentSchema, resolver_cache, Operation, Response
from ..utils import snake_case_to_camel_case

logger = logging.getLogger(__name__)
ResultType = TypeVar('ResultType', bound=Any)
AnswerType = TypeVar('AnswerType', bound=BaseAnswer)


@dataclass(frozen=True)
class Endpoint(typing.Generic[ResultType, AnswerType]):
    func: Callable[..., ResultType]  # request handler
    method: Literal['GET', 'POST', 'PUT', 'PATCH', 'DELETE', 'HEAD', 'OPTIONS', 'TRACE',
                    'get', 'post', 'put', 'patch', 'delete', 'head', 'options', 'trace']   # HTTP метод
    path: str  # HTTP path
    securities: tuple[SecurityRequirement, ...]  # possible ways to check auth, empty means "public"(no auth check)
    # maps handler parameter names to parameter objects, capable of producing value from incoming request
    arg_name_to_param: dict[str, CallParameter]
    answer_type: type[AnswerType]  # HTTP answer type
    # exception mapping to HTTP answer type, used to process <func> exceptions and make spec
    exception_type_to_answer_type: dict[type[Exception], type[Answer]]
    # HTTP answer type for unexpected exceptions
    unhandled_exception__answer_type: type[Answer]
    # main endpoint identification
    operation_id: str
    # Callable to convert <func> call result to an HTTP answer instance
    answer_factory: Callable[[ResultType], AnswerType]
    summary: str  # spec description
    description: str  # spec description
    tags: Iterable[str]  # mark endpoint with tags, spec only
    deprecated: bool  # mark endpoint as deprecated, spec only
    additional_parameters: tuple[SpecParameter, ...]  # add to spec, no other processing

    def __init__(
            self,
            func: Callable[..., ResultType],
            method: str,
            path: str = '',
            securities: Iterable[SecurityRequirement | SecurityScheme] = tuple(),
            arg_name_to_param: Mapping[str, CallParameter | type[RawBodyParameter]] = frozendict(),
            answer_type: type[AnswerType] = Answer[HttpStatusCode.OK],
            exception_type_to_answer_type: dict[type[Exception], type[Answer]] | None = None,
            unhandled_exception__answer_type: type[Answer] | None = ExceptionAnswer[HttpStatusCode.InternalServerError],
            operation_id: str = '',
            answer_factory: Callable[[ResultType], AnswerType] | None = None,
            summary: str | None = None,
            description: str | None = None,
            tags: list[str | Tag] | None = None,
            deprecated: bool = False,
            additional_parameters: tuple[SpecParameter, ...] | None = None,  # add to spec, no other processing
    ) -> None:
        error_text = 'path MUST be either specified directly(path="/test") or appended to "method"(method="GET/test")'
        error = ValueError(error_text + f" got {path=!r}, {method=!r}")
        if not path:
            raise_if('/' not in method, error)
            method, path = method.split("/", 1)
            path = '/' + path

        raise_if('/' in method or not path.startswith('/'), error)
        raise_if(not callable(func), ValueError(f'"func" arg value MUST be callable, got {func=!r}'))
        object.__setattr__(self, 'func', func)
        object.__setattr__(self, 'method', method)
        object.__setattr__(self, 'path', path)
        object.__setattr__(self, 'securities',
                           tuple(s.has() if isinstance(s, SecurityScheme) else s for s in securities))
        object.__setattr__(self, 'arg_name_to_param', {
            a: p() if isinstance(p, type) else p for a, p in arg_name_to_param.items()})
        object.__setattr__(self, 'answer_type', answer_type)
        object.__setattr__(self, 'exception_type_to_answer_type', exception_type_to_answer_type or {})
        object.__setattr__(self, 'unhandled_exception__answer_type', unhandled_exception__answer_type)
        object.__setattr__(self, 'operation_id', operation_id or func.__name__)
        object.__setattr__(self, 'answer_factory', answer_factory or answer_type)
        object.__setattr__(self, 'summary', summary or '')
        object.__setattr__(self, 'description', description or func.__doc__ or '')
        object.__setattr__(self, 'tags', tuple(tags or ()))
        object.__setattr__(self, 'deprecated', deprecated)
        object.__setattr__(self, 'additional_parameters', additional_parameters or tuple())
        get_securities_arg_names(self.securities)

    @cached_property
    def parameters(self) -> frozendict[CallParameter, None]:
        result: dict['SpecParameter', None] = {}
        for parameter in self.arg_name_to_param.values():
            result |= parameter.get_spec_parameters()
        if self.additional_parameters:
            for parameter in self.additional_parameters:
                result |= parameter.get_spec_parameters()
        return frozendict(result)

    @cached_property
    def body_parameters(self) -> tuple[BodyParameter, ...]:
        body_params: list[BodyParameter] = [param for param in self.parameters if isinstance(param, BodyParameter)]
        assert len(param_types := {type(param) for param in body_params}) <= 1, f"Can not combine {param_types} in body"
        assert len(param_mimes := {param.body_mime_type for param in body_params}) <= 1, \
            f"Can not combine {param_mimes} in body"
        return tuple(sorted(body_params, key=lambda x: x.default is not NoValue))

    @cached_property
    def body_parameter_type(self) -> type[BodyParameter] | None:
        return type(first(self.body_parameters, default=None))

    @cached_property
    def non_body_parameters(self) -> tuple[CallParameter, ...]:
        return tuple(param for param in self.parameters if not isinstance(param, BodyParameter))

    @cached_property
    def body_content_schema(self) -> type | None:
        if not (body_params := self.body_parameters):
            return None

        if self.body_parameter_type and issubclass(self.body_parameter_type, RawBodyParameter):
            return bytes

        class_: type = type(f'{snake_case_to_camel_case(self.operation_id)}Body', (EndpointBody,), {
            p.name: field(default=MISSING if p.default is not NoValue else p.default) for p in body_params})
        class_.__annotations__ = {param.name: param.annotated_schema for param in body_params}
        # noinspection PyTypeChecker
        return dataclasses.dataclass(kw_only=True)(class_)

    @cached_property
    def request_body(self) -> RequestBody | None:
        if (body_parameter_type := self.body_parameter_type) and (body_content_schema := self.body_content_schema):
            mime_type = body_parameter_type.body_mime_type
            content_types = mime_type if is_iterable(mime_type) else [mime_type]
            key_to_additional_param_info = {}
            if mime_type in (ContentType.MultipartForm, ContentType.FormUrlEncoded):
                for body_param in self.body_parameters:
                    if isinstance(body_param, MultipartParameter) and body_param.content_type or body_param.headers:
                        header_to_schema: dict[str, typing.Annotated | type] = {}
                        for body_part_header_param in body_param.headers:
                            param_name = body_part_header_param.name
                            param_schema = body_part_header_param.schema
                            raise_if((present_schema := header_to_schema.get(param_name, param_schema)) != param_schema,
                                     KeyError(f"Multipart header {param_name} is used with different schema in "
                                              f"multiple header parameters: {param_schema=}, {present_schema=}"))
                            header_to_schema[param_name] = param_schema

                        key_to_additional_param_info[body_param.name] = BodyPartDescription(
                            explode=body_param.explode, style=body_param.style,
                            content_type=body_param.allowed_content_type, headers=header_to_schema)
            return RequestBody({
                content_type: ContentSchema(body_content_schema, schema_key_to_description=key_to_additional_param_info)
                for content_type in content_types
            })

    # @cached_property
    # def code_to_answer(self) -> dict[http_tools.HttpStatusCode, type[http_tools.Answer]]:
    #     result = {self.answer_type.get_class_status_code(): self.answer_type}
    #     result |= {ans_t.get_class_status_code(): ans_t for ans_t in self.exception_type_to_answer_type.values()}
    #     if self.unhandled_exception__answer_type is not None:
    #         unhandled_exception__answer_code = self.unhandled_exception__answer_type.get_class_status_code()
    #         result[unhandled_exception__answer_code] = self.unhandled_exception__answer_type
    #     return result

    @property
    def _code_to_response(self) -> Mapping[int | Literal['default'], Response]:
        answer_types = [self.answer_type] + list(self.exception_type_to_answer_type.values())
        code_to_answer = {a.get_class_status_code(): a for a in answer_types}
        code_to_answer |= {'default': default_ans} if (default_ans := self.unhandled_exception__answer_type) else {}
        return frozendict({code: get_response_from_answer(answer) for code, answer in code_to_answer.items()})

    def as_operation(self) -> Operation:
        return Operation(
            path=self.path,
            method=self.method,
            operation_id=self.operation_id,
            securities=self.securities,
            parameters=[param for param in self.non_body_parameters if isinstance(param, SpecParameter)],
            request_body=self.request_body,
            code_to_response=self._code_to_response,
            summary=self.summary,
            description=self.description,
            tags=self.tags,
            deprecated=self.deprecated,
        )

    def gen_handler(self) -> Callable[[IncomingRequest], Awaitable[Answer | StreamAnswer]]:
        # TODO: check argument fullness
        if not self.securities:
            warnings.warn('Endpoints without securities will be prohibited')

        raise_if(
            inspect.isgeneratorfunction(self.func),
            NotImplementedError("Synchronous generator functions are not supported")
        )
        raise_if(
            inspect.isasyncgenfunction(self.func) and not issubclass(self.answer_type, StreamAnswer),
            TypeError("Endpoints, based on asynchronous generator function, require StreamResponse answer type")
        )

        async def handler(request: IncomingRequest) -> Answer | StreamAnswer:
            security_kwargs = None
            cache_token = resolver_cache.set({})
            security_key_to_exception = {}
            if self.securities:
                for security_requirement in self.securities:
                    try:
                        security_kwargs = await security_requirement.evaluate(request)
                        break
                    except* (Unauthorized, Forbidden) as exc_group:
                        if security_requirement.do_log:
                            security_key_to_exception[security_requirement.string_key] = ';'.join(
                                repr(e) for e in exc_group.exceptions)
            else:
                security_kwargs = {}
            resolver_cache.reset(cache_token)

            if security_kwargs is None:
                is_forbidden = any(isinstance(exc, Forbidden) for exc in security_key_to_exception.values())
                status_code = HttpStatusCode.Forbidden if is_forbidden else HttpStatusCode.Unauthorized
                # noinspection PyTypeChecker
                return JsonableAnswer(ErrorDescription(security_key_to_exception, "ErrorGroup"), status_code)

            try:
                param_kwargs = {
                    name: value for name, param in self.arg_name_to_param.items()
                    if (value := await param.get(request, security_kwargs)) is not NoValue
                }
                if inspect.isasyncgenfunction(self.func):
                    # noinspection PyTypeChecker
                    async_gen: collections.abc.AsyncGenerator = self.func(**(security_kwargs | param_kwargs))
                    first_result = await anext(async_gen)  # most possible exceptions will be raised here
                    stream_answer: StreamAnswer = self.answer_factory(first_result)
                    await stream_answer.prepare(request.request)
                    try:
                        async for value in async_gen:
                            await stream_answer.write(value)
                    except Exception as e:
                        logger.exception(e)
                    return stream_answer
                else:
                    call_result = await acall(self.func(**(security_kwargs | param_kwargs)))
                    return self.answer_factory(call_result)
            except Exception as e:
                for exception_type, answer_type in self.exception_type_to_answer_type.items():
                    if isinstance(e, exception_type):
                        return answer_type(e)

                if self.unhandled_exception__answer_type is not None:
                    traceback.print_exc()
                    logger.exception(e)
                    return self.unhandled_exception__answer_type(e)
                raise

        return handler

    # @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        parts = [
            f'func={self.func!r}', f'method={self.method!r}', f'path={self.path!r}',
            f'operation_id={self.operation_id!r}'
        ]
        parts += [f'securities={self.securities!r}'] if self.securities else []
        parts += [f'arg_name_to_param={dict(self.arg_name_to_param)!r}'] if self.arg_name_to_param else []

        parts += [f'request_body={self.request_body!r}'] if self.request_body is not None else []
        parts += [f'answer_type={repr_type(self.answer_type)}'] if self.answer_type else []
        if self.exception_type_to_answer_type:
            subparts = [f"{repr_type(e)}: {repr_type(a)}" for e, a in self.exception_type_to_answer_type.items()]
            parts += [f'exception_type_to_answer_type={{{", ".join(subparts)}}}']
        if self.exception_type_to_answer_type:
            parts += [f'unhandled_exception__answer_type={repr_type(self.unhandled_exception__answer_type)}']

        parts += [f'answer_factory={repr_type(self.answer_factory)}'] if self.answer_factory != self.answer_type else []
        parts += [f'summary={self.summary!r}'] if self.summary else []
        parts += [f'description={self.description!r}'] if self.description else []
        parts += [f'tags={list(self.tags)}'] if self.tags else []
        parts += [f'deprecated={self.deprecated}'] if self.deprecated else []
        parts += [f'additional_parameters={list(self.additional_parameters)}'] if self.additional_parameters else []
        return tuple(parts)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import functools
import itertools
from dataclasses import dataclass, field
from typing import Iterable, Mapping

from frozendict import frozendict
from init_helpers import Jsonable, AnyType

from .spec_ref import SpecRef
from .example import BaseExample, get_examples_as_dict
from .body_part_description import BodyPartDescription
from .spec_resource import SpecResource


@dataclass(frozen=True, slots=True)
class ContentSchema(SpecResource):
    """Media Type https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#media-type-object"""
    schema: AnyType
    examples: Iterable[BaseExample] = tuple()
    # multipart only
    schema_key_to_description: Mapping[str, BodyPartDescription] = field(default_factory=frozendict)

    def __post_init__(self):
        object.__setattr__(self, 'examples', tuple(self.examples))
        object.__setattr__(self, 'schema_key_to_description', frozendict(self.schema_key_to_description))

    def get_spec_dependencies(self) -> frozenset[SpecResource]:
        # noinspection PyTypeChecker
        return frozenset(itertools.chain([self.schema], self.examples, self.schema_key_to_description.values()))

    def get_spec_dict(self, dependency_to_ref: Mapping[SpecResource | AnyType, SpecRef]) -> frozendict[str, Jsonable]:
        result = {"schema": dependency_to_ref[self.schema]}
        result |= get_examples_as_dict(self.examples, dependency_to_ref)
        if self.schema_key_to_description:
            result["encoding"] = {k: dependency_to_ref[d] for k, d in self.schema_key_to_description.items()}
        return frozendict(result)

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        parts = [f'{self.schema.__name__ if isinstance(self.schema, type) else repr(self.schema)}']
        parts += [f'examples={self.examples!r}'] if self.examples else []
        if self.schema_key_to_description:
            parts += [f'schema_key_to_description={dict(self.schema_key_to_description)!r}']
        return tuple(parts)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from .any_of_schema import AnyOfSchema
from .all_of_schema import AllOfSchema
from .array_schema import ArraySchema
from .base_schema import BaseSchema
from .boolean_schema import BooleanSchema
from .build import SchemaBuilder
from .composite_schema import CompositeSchema
from .empty_schema import EmptySchema
from .integer_schema import IntegerSchema
from .null_schema import NullSchema
from .number_schema import NumberSchema
from .object_schema import ObjectSchema
from .schema_format import SchemaFormat, StringSchemaFormat, NumberSchemaFormat, IntegerSchemaFormat
from .string_schema import StringSchema
from .type_schema import TypeSchema



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from enum import StrEnum


class SecuritySchemeType(StrEnum):
    api_key = "apiKey"
    http = "http"
    oauth2 = "oauth2"
    open_id_connect = "openIdConnect"
    mutual_tls = "mutualTLS"



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import asyncio
import itertools
from dataclasses import dataclass
from typing import Callable, Any

from async_tools import acall
from frozendict import frozendict
from http_tools import IncomingRequest
from init_helpers import NoValue

from .parameter_aggregation import ParameterAggregation
from ..call_parameter import CallParameter
from ..spec_parameter import SpecParameter


@dataclass(frozen=True, kw_only=True)
class CallTemplate(ParameterAggregation):
    func: Callable
    args: tuple[CallParameter, ...]
    kwargs: dict[str, CallParameter]

    def __init__(self, func: Callable, *args: CallParameter, **kwargs: CallParameter):
        object.__setattr__(self, 'schema', NoValue)
        object.__setattr__(self, 'func', func)
        object.__setattr__(self, 'args', args)
        object.__setattr__(self, 'kwargs', kwargs)

    async def get(self, incoming_request: IncomingRequest, security_kwargs: dict[str, Any]) -> Any:
        args = await asyncio.gather(*[arg.get(incoming_request, security_kwargs) for arg in self.args])
        kwarg_values = await asyncio.gather(*[v.get(incoming_request, security_kwargs) for v in self.kwargs.values()])
        args = [value for value in args if value is not NoValue]
        kwargs = {key: value for key, value in zip(kwarg_values, self.kwargs) if value is not NoValue}
        result = await acall(self.func(*args, **kwargs))
        return result

    def get_spec_parameters(self) -> frozendict['SpecParameter', None]:
        result = {}
        for arg in itertools.chain(self.args, self.kwargs.values()):
            result |= arg.get_spec_parameters()
        return frozendict(result)



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
import functools
from dataclasses import dataclass
from typing import Mapping

from frozendict import frozendict
from init_helpers import Jsonable

from .spec_ref import SpecRef


@dataclass(frozen=True, slots=True)
class SpecResource(abc.ABC):
    @abc.abstractmethod
    def get_spec_dependencies(self) -> frozenset['SpecResource']:
        raise NotImplementedError

    @functools.cache
    def get_spec_dependencies_recursively(self) -> frozenset['SpecResource']:
        visited: set['SpecResource'] = set()
        pending = list(self.get_spec_dependencies())
        while pending:
            current = pending.pop()
            if current in visited:
                continue
            visited.add(current)
            pending.extend(current.get_spec_dependencies())
        return frozenset(visited)

    @abc.abstractmethod
    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef]) -> frozendict[str, Jsonable]:
        raise NotImplementedError

    def merge_spec_dict(
            self,
            dependency_to_ref: Mapping['SpecResource', SpecRef],
            present_value: frozendict[str, Jsonable]
    ) -> frozendict[str, Jsonable]:
        raise TypeError(f'{self!r} cannot be merged with {present_value!r}')



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import functools
from dataclasses import dataclass
from typing import Mapping

from frozendict import frozendict
from http_tools import ContentType
from init_helpers import Jsonable

from .content_schema import ContentSchema
from .spec_ref import SpecRef
from .spec_resource import SpecResource


@dataclass(frozen=True, slots=True)
class RequestBody(SpecResource):
    """Request Body https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.1.0.md#request-body-object"""
    content_type_to_schema: Mapping[ContentType, ContentSchema]
    description: str = ''
    required: bool = True

    def __post_init__(self):
        object.__setattr__(self, 'content_type_to_schema', frozendict(self.content_type_to_schema))

    def get_spec_dependencies(self) -> frozenset['SpecResource']:
        return frozenset(schema for schema in self.content_type_to_schema.values())

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef]) -> frozendict[str, Jsonable]:
        result: dict[str, Jsonable] = {'required': self.required} if self.required else {}
        # noinspection PyTypeChecker
        result['content'] = frozendict({t: dependency_to_ref[s] for t, s in self.content_type_to_schema.items()})
        result |= {'description': self.description} if self.description else {}
        return frozendict(result)

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        parts = [repr({content_type.value: schema for content_type, schema in self.content_type_to_schema.items()})]
        parts += [f'description={self.description!r}'] if self.description else []
        parts += [f'required={self.required!r}'] if not self.required else []
        return tuple(parts)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
def snake_case_to_camel_case(snake_case_string: str, first_camel: bool = True) -> str:
    """
    Converts a snake_case string to camelCase.
    """
    words = snake_case_string.split('_')
    if first_camel:
        return ''.join(word.capitalize() for word in words)
    else:
        return words[0] + ''.join(word.capitalize() for word in words[1:])



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
import contextvars
from dataclasses import dataclass, field
from typing import TypeAlias, TypeVar, Callable, Awaitable, Iterable, Literal, Any, Mapping, TYPE_CHECKING, Generic, \
    ClassVar

from async_tools import acall
from frozendict import frozendict
from http_tools import IncomingRequest

from init_helpers import DataclassProtocol, raise_if, Jsonable

from openapi_tools.spec import SpecResource, SpecRef, ReferencableResource
from ..exceptions import Forbidden
from .security_scheme_type import SecuritySchemeType

if TYPE_CHECKING:
    from ..requirement import SecurityRequirement

AuthToken: TypeAlias = str
AuthInfo = TypeVar("AuthInfo", bound=dict | DataclassProtocol)
resolver_cache = contextvars.ContextVar('resolver_cache')


@dataclass(frozen=True, slots=True, repr=False)
class SecurityScheme(Generic[AuthInfo], ReferencableResource, abc.ABC):
    can_be_inlined: ClassVar[bool] = False
    spec_path: ClassVar[tuple[str, ...]] = ('components', 'security')
    type_: ClassVar[SecuritySchemeType]
    resolver: Callable[[AuthToken], Awaitable[tuple[AuthInfo, Iterable[str]]] | tuple[AuthInfo, Iterable[str]]]
    do_log: bool = field(default=True, kw_only=True)

    @classmethod
    def get_spec_dependencies(cls) -> frozenset['SpecResource']:
        return frozenset()

    @abc.abstractmethod
    def get_spec_dict(self, dependency_to_ref: Mapping[SpecResource, SpecRef]) -> frozendict[str, Jsonable]:
        return frozendict({'type': self.type_})

    def get_spec_ref(self, key: str = '') -> str:
        return key if key else self.get_key()

    @abc.abstractmethod
    async def _extract_token(self, incoming_request: IncomingRequest) -> AuthToken:
        pass

    async def evaluate(self, incoming_request: IncomingRequest, required_scopes: frozenset[str]) -> AuthInfo:
        cache = resolver_cache.get({})
        if self not in cache:
            auth_token = await self._extract_token(incoming_request)
            try:
                cache[self] = await acall(self.resolver(auth_token))
            except Exception as e:
                cache[self] = e
        result = cache[self]
        if isinstance(result, Exception):
            raise result
        auth_info, allowed_scopes = cache[self]
        allowed_scopes = allowed_scopes if isinstance(allowed_scopes, set) else set(allowed_scopes)
        raise_if(missing_scopes := required_scopes - allowed_scopes,
                 Forbidden(f'missing scopes: {", ".join(sorted(missing_scopes))}'))
        return auth_info

    # @property
    # @abc.abstractmethod
    # def key(self) -> str:
    #     pass

    def has(self, *args: str, **kwargs: type[Literal] | Callable[[AuthInfo], Any]) -> 'SecurityRequirement':
        from ..requirement import SecurityRequirement
        return SecurityRequirement({self: args}).use(**kwargs)

    def _repr(self) -> list[str]:
        return [f"resolver={self.resolver!r}"]

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._repr())})'

    __str__ = __repr__



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from typing import Iterable

from init_helpers import raise_if

from .requirement import SecurityRequirement
from .scheme import SecurityScheme


def get_securities_arg_names(securities: Iterable[SecurityRequirement | SecurityScheme]) -> frozenset[str]:
    result = None
    for security in securities:
        names = set(security.argument_name_to_getter.keys()) if isinstance(security, SecurityRequirement) else set()
        if result is None:
            result = names
        else:
            raise_if(names != result, ValueError(f'All {securities=} must provide same arguments: {names} != {result}'))
    return frozenset(result or tuple())



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import functools
from abc import ABC
from dataclasses import dataclass
from typing import Mapping, Iterable, ClassVar

from frozendict import frozendict
from init_helpers import NoValue, Jsonable

from openapi_tools.spec.spec_resource import SpecResource, SpecRef
from .base_schema import BaseSchema


@dataclass(frozen=True, slots=True)
class CompositeSchema(BaseSchema, ABC):
    composition_keyword: ClassVar[str]
    variants: Iterable[BaseSchema]
    discriminator_property_name: str = ''
    discriminator_value_to_schema: Mapping[Jsonable, BaseSchema] | None = None

    def __post_init__(self):
        object.__setattr__(self, 'variants', tuple(self.variants))

    @property
    def has_default(self) -> bool:
        return self.default is not NoValue or any(v for v in self.variants if v.default is not NoValue)

    @functools.cache
    def get_spec_dependencies(self) -> frozenset['SpecResource']:
        return frozenset(BaseSchema.get_spec_dependencies(self) | set(self.variants))

    def get_spec_dict(self, dependency_to_ref: Mapping['SpecResource', SpecRef | dict]) -> frozendict[str, Jsonable]:
        result = {self.composition_keyword: [dependency_to_ref[option] for option in self.variants]}
        if self.discriminator_property_name:
            if set(self.discriminator_value_to_schema.values()) != set(self.variants):
                raise ValueError(f'{self.discriminator_value_to_schema=} must have all values of {self.variants=}')
            result['discriminator'] = {
                'propertyName': self.discriminator_property_name,
                'mapping': {value: dependency_to_ref[schema] for value, schema in self.discriminator_value_to_schema}
            }
        return frozendict(BaseSchema.get_spec_dict(self, dependency_to_ref) | result)

    @functools.cache
    def _get_repr_parts(self) -> tuple[str, ...]:
        parts = list(BaseSchema._get_repr_parts(self))
        parts += [f'{self.variants!r}'] if self.variants else []
        if self.discriminator_property_name:
            parts.append(f'discriminator_property_name={self.discriminator_property_name!r}')
        return tuple(parts)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(self._get_repr_parts())})'

    __str__ = __repr__
